/*########################################################################
#    Copyright (C) 2014  Bruno Ribeiro
#
#Permission is hereby granted, without written agreement and without
#license or royalty fees, to use, copy, modify, and distribute this
#software and its documentation for any purpose, provided that the
#above copyright notice and the following two paragraphs appear in
#all copies of this software.
#
#IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
#DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
#ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
#IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
#DAMAGE.
#
#THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
#BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
#FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
#ON AN "AS IS" BASIS, AND THE COPYRIGHT HOLDER HAS NO OBLIGATION TO
#PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
########################################################################
*/
#include <iostream>
#include <string>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <inttypes.h>
#include <limits.h>
#include <cmath>
#include <stack>
#include <stdio.h>
#include "RandomLib/Random.hpp"

#define _FLOAT double
#define LONGFL long double
#define _PTYPE float
#define _FLAGTYPE uint8_t

#define VERTEXSAMPLING 1
#define FRONTIERSAMPLING 2
#define EDGESAMPLING 3
#define RWSAMPLING 5
#define SNOWBALLSAMPLING 4

//largest positive value an int32_t can hold.
#define INT32_MAX 0x7fffffffL

//largest negative value an int32_t can hold.
#define INT32_MIN (-INT32_MAX - 1L) 


// 		//Returns the value ln[(xx)] for xx > 0.
// 		FLOAT gammln(_FLOAT xx)
// 		{
	// 			//    Internal arithmetic will be done in FLOAT precision, a nicety that you can omit if five-figure accuracy is good enough.
	// 			LONGFL x,y,tmp,ser;
	// 			static LONGFL cof[6]={76.18009172947146,-86.50532032941677,
	// 			24.01409824083091,-1.231739572450155,
	// 			0.1208650973866179e-2,-0.5395239384953e-5};
	// 			int j;
	// 			
	// 			y=x=xx;
	// 			tmp=x+5.5;
	// 			tmp -= (x+0.5)*logl(tmp);
	// 			ser=1.000000000190015;
	// 			for (j=0;j<=5;j++) ser += cof[j]/++y;
	// 			return -tmp+logl(2.5066282746310005*ser/x);
	// 		}
	// 		
	// 		
	// 		_FLOAT factln(INT n)
	// 		//Returns ln(n!).
	// 		{
		// 			//             A static array is automatically initialized to zero.
		// 			static FLOAT a[101];
		// 			if (n < 0) return -1e90;
		// 			if (n <= 1) return 0.0;
		// 			//            In range of table.
		// 			if (n <= 100) return a[n] ? a[n] : (a[n]=gammln(n+1.0));
		// 			//            Out of range of table.
		// 			else return gammln(n+1.0);
		// 		}
		// 		
		// 		
		// 		_FLOAT bico(INT n, INT k)
		// 		//Returns the binomial coefficient (n,k) as a floating-point number.
		// 		{
			// 			return floor(0.5+EXPF(factln(n)-factln(k)-factln(n-k)));
			// 			//    The floor function cleans up roundoff error for smaller values of n and k.
			// 		}
			
using namespace std;

RandomLib::Random rnd;		// r created with random seed

inline int myrand(int max) {
	return rnd.IntegerC(0,max-1);
}

//#include <stdio>

#define MIN(a,b) ((a)>(b)?(b):(a))
#define MAX(a,b) ((a)<(b)?(b):(a))
#define ABSMAX(a,b) (abs(a)<abs(b)?(b):(a))

#define MAXGTOHIST (10)
#define NOPRINTORIG
//#define DEBUGFRONT
//#define TESTDIST


struct edge {
  int32_t u,v;
  _PTYPE p;
};

int compare_edges(const void *e1, const void *e2){
    return ((*((edge **)e1))->v - (*((edge **)e2))->v);
}

struct orderededge {
  edge *e;
  int32_t order;
};


struct vertex {
	int32_t  deg;
	int32_t  indeg, outdeg;
	int32_t  neigsize;
    int32_t id;
	_FLAGTYPE flag;
	edge** edges;
};

void *zrealloc(void *ptr, size_t newsize, size_t oldsize) {
  int8_t *p = (int8_t *) realloc(ptr,newsize);
  if (p == NULL) {
      cerr << "Could not allocate "<<((_FLOAT)(newsize))/(1024*1024)<<"Kbytes. Aborting."<<endl;
      abort();
  }
  memset(p+oldsize,0,newsize - oldsize);
  return (void*)p;
}

void *zalloc(size_t newsize) {
  int8_t *p = (int8_t *) malloc(newsize);
  if (p == NULL) {
      cerr << "Could not allocate "<<((_FLOAT)(newsize))/(1024*1024)<<"Kbytes. Aborting."<<endl;
      abort();
  }
  memset(p,0,newsize);
  return (void*)p;
}

class Gdb {
private:
  int32_t reservedspaceE;
  int32_t reservedspaceV;
  int32_t edgeindex;
  int32_t N;
  int32_t _MAXID;
  int32_t _MAXDEGREE;
  int32_t noedges;
  
  int32_t i_idE, i_idV;
  
  vertex *V;
  edge *E;

  void updatevertex(int32_t u) {
//    cerr << "add="<<pe<<endl;
    if (u >= reservedspaceV) {
      cerr << "Fatal Error: More vertices than previously allocated = "<<reservedspaceV<<endl;
      abort();
    }
    _MAXID = MAX(_MAXID,u);
    if (V[u].edges == NULL) {
      N++;
      V[u].neigsize = 4;
      V[u].edges = (edge **) zalloc(V[u].neigsize*sizeof(edge*));
    }
    if (V[u].deg >= V[u].neigsize) {
      V[u].edges = (edge **) zrealloc(V[u].edges, (V[u].neigsize*2)*sizeof(edge**), V[u].neigsize*sizeof(edge**));
      V[u].neigsize *= 2;
    }
    V[u].id = u;
  }
  
public:
  Gdb() {
  }

  void init(int32_t p_reservedspaceE, int32_t p_reservedspaceV) {
    edgeindex = 0;
    N = 0;
    _MAXDEGREE = 0;
    _MAXID = 0;
	noedges = 0;
    reservedspaceE = p_reservedspaceE;
    reservedspaceV = p_reservedspaceV;
    E = (edge *) zalloc(reservedspaceE * sizeof(edge));
    V = (vertex *) zalloc(reservedspaceV * sizeof(vertex));
  }
  
  inline _FLOAT samplingFreq(vertex *) {
	  return 1.0/((_FLOAT)size());
  }
  
  inline vertex* getvertex(int32_t u) {
    return (exists(u))? &V[u] : NULL;
  }
  
  inline void edge_iterator_start() {
    i_idE = 0;
  }
  
  inline bool exists_e(int32_t id) {
	  return ((id < edgeindex)? ((E[id].v >= 0)? true : false ): false);
  }
  
  inline bool exists(edge *e) {
	  return ((e != NULL)? ((e->v >= 0)? true : false ): false);
  }
  
  edge* next_edge() {
	  while ((!exists_e(i_idE)) && (i_idE < edgeindex)) {
		  i_idE++;
	  }
	  return (exists_e(i_idE)? &E[i_idE++] : NULL );
  }
  
  inline void vertex_iterator_start() {
    i_idV = 0;
  }
  
  vertex* next_vertex() {
    while ((!exists(i_idV)) && (i_idV <= MAXID()))
      i_idV++;
    if (i_idV > MAXID())
      return NULL;
    return &V[i_idV++];
  }
  
  inline edge* getedge(int32_t id) {
	  return (exists(&E[id])? &E[id] : NULL);
  }
  
/*  vertex* getneighbor(int32_t u, int32_t k) {
    if (exists(u)) {
      if (k >= V[u].deg)
        return NULL;
      return &V[(V[u].edges[k])->v];
    }
    return NULL;
  }
  */
  vertex* getneighbor(vertex *u, int32_t k) {
    if (u != NULL) {
		if (!exists(u->edges[k])) return NULL;
		return &V[(u->edges[k])->v];
    }
    return NULL;
  }
  
  edge* getedge(vertex *u, int32_t k) {
    if (u != NULL) {
		if (!exists(u->edges[k])) return NULL;
		return u->edges[k];
    }
    return NULL;
  }
  
  vertex* getrandomneighbor(vertex *u) {
    int32_t i;
    if (u != NULL) {
      if (u->deg == 0) return u;
      i = myrand(u->deg);
	  if (!exists(u->edges[i])) return NULL;
	  return &V[(u->edges[i])->v];
    }
    return NULL;
  }
  
  edge* randomedge() {
	  int32_t r;
	  
	  do {
		  r = myrand(edgeindex);
	  } while (!exists(&E[r]));
	  return &E[r];
  }
  
  edge* randomedge(vertex *u) {
	  edge *e;
	  if (u != NULL) {
		  if (u->deg > 0) {
			  do {
				  e = u->edges[myrand(u->deg)];
			  } while (!exists(e));
			  return e;
		  }
	  }
	  return NULL;
  }
  
  vertex* randomvertex() {
    int u = (myrand(_MAXID+1));
    while (!exists(u)) 
      u = (myrand(_MAXID+1));
    return &V[u];
  }
  
  inline int32_t indegree(vertex *u) {
	  return ((u != NULL)? (u->indeg) : (-1));
  }
  
  inline int32_t outdegree(vertex *u) {
	  return ((u != NULL)? (u->outdeg) : (-1));
  }
  
  inline int32_t degree(vertex *u) {
	  return ((u != NULL)? (u->deg) : (-1));
  }
  
  inline int32_t degree(int32_t u) {
	  return degree(getvertex(u));
  }
  
  inline int32_t indegree(int32_t u) {
	  return indegree(getvertex(u));
  }
  
  inline int32_t outdegree(int32_t u) {
	  return outdegree(getvertex(u));
  }
  
  inline int32_t MAXID() {
    return _MAXID;
  }

/*  inline int32_t MAXgValue() {
    return _MAXgValue;
  }
*/

  inline int32_t MAXDEGREE() {
    return _MAXDEGREE;
  }

  inline int32_t size() {
    return N;
  }

  inline int32_t no_edges() {
    return noedges;
  }

  inline _FLAGTYPE flag(int32_t n) {
    return ( exists(n)? (V[n].flag) : (-1)  );
  }

  inline _FLAGTYPE flag(vertex *v) {
	  return ( exists(v)? (v->flag) : (-1)  );
  }
  
	inline void setflag(int32_t n, _FLAGTYPE flag) {
		if (exists(n)) {
			V[n].flag = flag;
		}
	}

	inline void setflag(vertex *v, _FLAGTYPE flag) {
		if (exists(v)) {
			v->flag = flag;
		}
	}

   void zeroflags() {
	zeroflags(~0x00);
   }

  void zeroflags(int8_t mask) {
	int32_t i;
	for (i=0; i <= MAXID(); i++) {
		setflag(i,(flag(i) & (~mask)));
	}
  }

   void zeroedgep() {
    int32_t i;
	for (i=0; i < edgeindex; i++) {
      E[i].p = 0;
    }
  }

   void setAlledgep(_FLOAT p) {
    int32_t i;
	for (i=0; i < edgeindex; i++) {
      E[i].p = p;
    }
  }

  inline bool exists(int32_t n) {
    return ((n <= MAXID())? (V[n].edges != NULL) : false);
  }
  
  inline bool exists(vertex *v) {
	  return (v != NULL);
  }
  
  edge * reverseEdge(edge *e) {
	  vertex *v;
	  int32_t k;
	  
	  if (exists(e)) {
		  v = getvertex(e->v);
		  if (exists(v)) {
			  for (k =0; k < v->deg; k++) {
				  if (v->edges[k]->v == e->u) 
					  return v->edges[k];
			  }
		  }
	  }
	  return NULL;
  }
  
  void trimGraphDifferent(_FLAGTYPE flag) {
	  int32_t i,j;
	  int32_t N_MAXDEGREE = 0;
	  int32_t N_MAXID = 0;

	  for (i=0; i <= MAXID(); i++) {
		  if (exists(i)) {
			  if (V[i].flag != flag) {
				  // edge is removed
				  N--;
				  if (V[i].edges != NULL) {
					  for (j = 0; j < V[i].deg; j++) {
						  if (exists(V[i].edges[j])) {
							  noedges--;
							  V[i].edges[j]->u = -1;
							  V[i].edges[j]->v = -1;
						  }
					  }
					  free(V[i].edges);
					  V[i].edges = NULL;
				  }
				  V[i].deg = 0;
			  } else {
				  // edge stays
				  N_MAXDEGREE = MAX(N_MAXDEGREE,V[i].deg);
				  N_MAXID = MAX(N_MAXID,V[i].id);
			  }
		  }
	  }
	  _MAXDEGREE = N_MAXDEGREE;
	  _MAXID = N_MAXID;
  }
  

  void addedge(int32_t u, int32_t v, _PTYPE p, bool trueedge) {
    vertex *up;
    int32_t k;
    
    if (edgeindex >= reservedspaceE) {
      cerr << "Fatal Error: More edges than previously allocated = " << reservedspaceE << endl;
      abort();
    }
    // see if edge is repeated
    up = getvertex(u);
    if (up != NULL) {
      for (k=0; k < up->deg; k++) {
        if (up->edges[k] != NULL) {
//         cerr << "q " << up->edges[k] << "  deg = "<<up->deg<<endl;
          if (up->edges[k]->v == v) {
//            repeated edge... get out
				if (trueedge) V[u].outdeg++;
				if (trueedge) V[v].indeg++;
				if (trueedge) up->edges[k]->p = p;
				return;
          }
        }
      }
    }
    
    E[edgeindex].u = u;
    E[edgeindex].v = v;
    E[edgeindex].p = p;
	updatevertex(u);
	updatevertex(v);
	V[u].edges[V[u].deg] = &E[edgeindex];
	V[u].deg++;
	_MAXDEGREE = MAX(_MAXDEGREE,V[u].deg);
	if (trueedge) V[u].outdeg++;
	if (trueedge) V[v].indeg++;
	V[u].flag = 0;
    edgeindex++;
	noedges++;
  }
  
  void print_edge_probabilities() {
	  int32_t i;

	  for (i=0; i < edgeindex; i++) {
		  cerr<< E[i].u <<"\t"<< E[i].v << "\t" << E[i].p << endl;
	  }
  }
  
};


class Graph : public Gdb {

  private:

    void read_file(const char *file)
    {
		FILE *f;
        vertex *v;
		int32_t i,j;
        int32_t maxid = 0;
        int32_t p_reservedspaceE;
		char st[1000];
		
        p_reservedspaceE = 0;
		sprintf(st,"zcat %s",file);
		f = popen(st,"r");
		if (f == NULL) {
			cerr << "# Error opening file "<<file<<endl;
			exit(1);
		}
		while (!feof(f)) {
		  if (fscanf(f,"%d\t%d\n",&i,&j) == 2){
			  if (i != j) {
				  maxid = MAX(MAX(maxid,i),j);
				  p_reservedspaceE+=2;
			  }
		  }
		}
        pclose(f);
		if (maxid == 0) {
			cerr << "# Error with graph file "<<file<<endl;
			exit(1);
		}
        cerr << "#Edges = "<<p_reservedspaceE<<endl;
        cerr << "#Vertices = "<<maxid+1<<endl;
        init(p_reservedspaceE,maxid+1);
        f = popen(st,"r");
        while (!feof(f)) {
          if (fscanf(f,"%d\t%d\n",&i,&j) == 2){
			  if (i != j) {
				addedge(i,j,1,true);
				// creates edge j->i (because the graph is directed! and we need an undirected graph)
				addedge(j,i,0,false);
			  }
          }
        }
        cerr <<"#Maximum degree = "<<MAXDEGREE()<<endl;
        pclose(f);
        
        vertex_iterator_start();
        while ((v = next_vertex())) {
            qsort(v->edges, v->deg,sizeof(edge *),compare_edges);
        }

	}
 
  public: 
	
	Graph(const char *filename) 
    :Gdb()
    { 
	  read_file(filename);
	}
	
	~Graph() {
	}
    
};

int32_t compare_ints (const void *a, const void *b) {
  if (*((int32_t*)a) > *((int32_t*)b))
	return 1;
  else if (*((int32_t*)a) < *((int32_t*)b))
	return -1;
  else
	return 0;
}	

class NodeSet {
  private:
    Graph *G;
    vertex **vertices;
    int32_t ndidx;
    int32_t reserved;
    int32_t totaldeg;
	int32_t it;
    
    
  public:
    NodeSet(Graph *pG) {
      G = pG;
      reserved = 10000;
      vertices = (vertex **) zalloc(sizeof(vertex *)*reserved);
      ndidx = 0;
      totaldeg = 0;
    }
    
    ~NodeSet() {
      free(vertices);
    }
    
    void clear(){
      totaldeg = 0;
      ndidx = 0;
    }
	
	inline void iterator_start() {
		it = 0;
	}
	
	inline vertex *next_vertex() {
		return element(it++);
	}
    
    void add(vertex *u) {
      if (ndidx >= reserved) {
        vertices = (vertex **) zrealloc(vertices,sizeof(vertex *)*(reserved+10000),sizeof(vertex *)*reserved);
        reserved += 10000;
      }
      vertices[ndidx++] = u;
      totaldeg += G->degree(u);
    }
	
	void eraseall(vertex *u) {
		int32_t i,k;
		
		for (i=0; i < ndidx; i++) {
			if (vertices[i] == u) {
				ndidx--;
				totaldeg -= G->degree(u);
				for (k=i; k < ndidx; k++) {
					vertices[k] = vertices[k+1];
				}
				vertices[ndidx] = 0;
			}
		}
	}
	
    void eraseone(vertex *u) {
      int32_t i,k;
      
      for (i=0; i < ndidx; i++) {
        if (vertices[i] == u) {
          ndidx--;
          totaldeg -= G->degree(u);
          for (k=i; k < ndidx; k++) {
            vertices[k] = vertices[k+1];
          }
          vertices[ndidx] = 0;
		  break; // should remove just one
        }
      }
    }
    
    vertex* element(int32_t k) {
      if (k >= size()) {
        return NULL;
      }
      return vertices[k];
    }
    
    vertex* randomvertex_fromedge() {
      int32_t e,i;
      
      e = (myrand(totaldeg)) + 1;
//      cerr << "Chosen edge "<< e << " from " << totaldeg << " edges" << endl;
      i = 0;
      while (e > 0) {
        e -= G->degree(vertices[i++]);
      }
//      cerr << "Chosen vertex "<< vertices[i-1]->id << " with degree " <<  vertices[i-1]->deg << endl;
      return vertices[i-1];
    }
    
/*    void freq(NodeStatistics *freq) {
      int32_t i;
      
      freq->clear();
      for (i=0; i < idx; i++) {
        freq->inc(G->g(vertices[i]));
      }
    }
    */
/*    void print(const string str) {
      int32_t i;
      
      for (i=0; i < size(); i++) {
        cerr << str << "\t" << G->g(vertices[i]) << endl;
      }
    }
    */
	inline int32_t size() {
		return ndidx;
	}

	inline int32_t no_edges() {
		return totaldeg;
	}
};

class EdgeSet {
  private:
    Graph *G;
    int32_t reservedspace;
    int32_t edgeindex;
	int32_t it;
    orderededge *edges;
    
	inline int32_t space_left() {
		return (reservedspace - edgeindex);
	}
	
	public:
    
    EdgeSet(Graph *pG) {
      G = pG;
      // Vector with all vertex ids that were uniformly sampled
      reservedspace = 10000;
      edges = (orderededge *) zalloc(sizeof(orderededge)*reservedspace);
      edgeindex = 0;
    }
    
    ~EdgeSet() {
      free(edges);
    }
    
    void clear(){
      edgeindex = 0;
    }
    
//     void print(char *str) {
//       int32_t i;
//       
//       for (i = 0; i < edgeindex; i++) {
//         cerr << str << "\t" << G->g(G->getvertex(edges[i].e->u)) << "\t" << G->g(G->getvertex(edges[i].e->v)) << "\t" << edges[i].order << endl;
//       }
//     }
// 	
	void iterator_start() {
		it = 0;
	}
	
	edge *next_edge() {
		if (it != edges[it].order) {
			cerr << "Wrong edge order, it = "<< it << " and edge order = "<<edges[it].order<<endl;
		}
		return (edges[it++].e);
	}
    
    
    void add(edge *e, int32_t order) {
      if (space_left() <= 0) {
        edges = (orderededge *) zrealloc(edges,sizeof(orderededge)*(reservedspace+1000),sizeof(orderededge)*reservedspace);
        reservedspace += 1000;
      }
      edges[edgeindex].e = e;
      edges[edgeindex].order = order;
      edgeindex++;
    }
    
    orderededge* element(int32_t k) {
      if (k < size()) {
        return &edges[k];
      }
      return NULL;
    }

    inline int32_t size() {
      return edgeindex;
    }
    
};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------
class F {
	private:
		Graph *Gr;
		
	public:
		F(Graph *pG) {
			Gr = pG;
		}
		
		virtual ~F() {
		}
		
		inline Graph *G() {
			return Gr;
		}
		
		virtual void reset() {
		}
		
		virtual void prepare_nextrun() {
		}
		
		virtual void add(edge *, int32_t , int8_t, _FLOAT ) {
		}
			
		virtual void  consolidate_TrueSum() {
		}
		
		virtual _FLOAT normalize_val(_FLOAT *) {
			return 0.0;
		}
		
		virtual _FLOAT normalize_TrueSum() {
			return 0.0;
		}
		
		virtual  _FLOAT* trueSum() {
			return NULL;
		}
		
		virtual void consolidate() {
		}
		
		virtual void print(const char *, F *) {
		}
};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------
class FrontierSampling {
private:
	Graph *G;
	
public:
	FrontierSampling(Graph *pG) {
		G = pG;
	}

	_FLOAT runsampled(int64_t samplesbgt, int64_t front_size, int64_t RUNS, bool VarDist, F *Fsum, int64_t *sedges){
		NodeSet *randomNodes;
		NodeSet *samplednodes;
		int64_t runs;
		int64_t samplebudget;
		int64_t sampled;
		int64_t k;
		vertex *r;
		_FLOAT err = -1;
		edge *e;
		vertex *u,*v;
		_FLOAT avgdeg;
		
		#ifdef TESTDIST
		int64_t t;
		int32_t degdist[100];
		memset(degdist,0,100*sizeof(int32_t));
		#endif

		randomNodes = new NodeSet(G);
		samplednodes = new NodeSet(G);
		
		*sedges = 0;
		if (Fsum != NULL) Fsum->reset();

		for (runs = 0; runs < RUNS; runs++) {
			samplebudget = samplesbgt;
			randomNodes->clear();
			samplednodes->clear();
			if (Fsum != NULL) Fsum->prepare_nextrun();
			avgdeg = 0.0;
			for (k =0; k < front_size; k++) {
				// node has to be outside the frontier...
				r = G->randomvertex();
				//      cerr << "random node: "<<r<<endl;
				#ifdef DEBUGFRONT
					cerr << "random vertex " << r->id << endl;
				#endif
				randomNodes->add(r);
				samplednodes->add(r);
				avgdeg += G->degree(r);
				if (Fsum != NULL) {
					e = G->getedge(r,0); // pass any edge as we are just interested in e->u
					Fsum->add(e,-1,VERTEXSAMPLING,0.0);
				}
				#ifdef TESTDIST
				//        degdist[r->deg]++;
				#endif
			}
			avgdeg /= front_size;
			sampled = front_size;
			e = NULL;
			v = NULL;
			
			while (sampled < samplebudget) {
			// randomly sample edges
				u = samplednodes->randomvertex_fromedge();
				samplednodes->eraseone(u);
				e = G->randomedge(u);
				v = G->getvertex(e->v);
			#ifdef DEBUGFRONT
				cerr << "Edge sampled: "<< u->id << "," << v->id<< " sampled so far "<<sampled<<endl;
			#endif
				if (Fsum != NULL) {
					if (sampled > front_size) // do not sampled the first edge (it is unecessarely biased if edge degrees are independent)
						Fsum->add(e,sampled-front_size,FRONTIERSAMPLING,0.0);
				}
				sampled++;
				samplednodes->add(v);
			}
			#ifdef TESTDIST
			if (v != NULL) degdist[v->deg]++;
			#endif
			if (Fsum != NULL) Fsum->consolidate();
		}

		
		
		#ifdef TESTDIST
		cerr << "Dist = " << endl;
		t =0;
		for (k = 0; k < 14; k++) {
			t+=degdist[k];
		}
		for (k = 0; k < 14; k++) {
			cerr << "   d[" << k << "] = " << (_FLOAT)degdist[k]/t << endl;
		}
		#endif

		delete randomNodes;
		delete samplednodes;

		return err;
	}

};

class RandomWalkSampling {
 private:
  Graph *G;
  
 public:
  RandomWalkSampling(Graph *pG) {
    G = pG;
  }
  
  _FLOAT runsampled(int64_t samplebudget, int32_t independentWalkers, int64_t RUNS, bool VarDist, F *Fsum, int64_t *sedges) {
    vertex *r;
    edge *e;
//	edge *old_e;
    int64_t sampled, walker, runs;
    _FLOAT err = 0;
	
	*sedges = 0;
//	old_e = NULL;
    
	if (Fsum != NULL) Fsum->reset();
	
	if (samplebudget % independentWalkers != 0) {
      cerr << "Warning in multiple random walks: Sample budget must be a multiple of the random walkers";
    }
    
    for (runs = 0; runs < RUNS; runs ++ ) {
	  if (Fsum != NULL) Fsum->prepare_nextrun();
      for (walker = 0; walker < independentWalkers; walker++) {
        r = G->randomvertex();
        sampled = 0;
        while (sampled < samplebudget/independentWalkers) {
		  e = G->randomedge(r);
          r = G->getvertex(e->v);
		  if (Fsum != NULL) {
	//		  if (sampled > 0) // avoids sampling the first edge because it may be too biased
				if (sampled == 0)
					Fsum->add(e,sampled,VERTEXSAMPLING,G->degree(e->u));
				else
					Fsum->add(e,sampled,RWSAMPLING,0.0);
		  }
          sampled++;
        }
      }
	  if (Fsum != NULL) Fsum->consolidate();
	}
	return err;
  }
};

class RandomEdgeSampling {
 private:
  Graph *G;
 
 public:
  RandomEdgeSampling(Graph *pG) {
    G = pG;
  }

  _FLOAT runsampled(int64_t samples, int32_t runs, bool VarDist, F *Fsum) {
    int64_t i,k;
    edge *e;
	_FLOAT err = 0.0;
	
//	G->zeroedgep();
	if (Fsum != NULL) Fsum->reset();
	
	for (k=0; k < runs; k++) {
		if (Fsum != NULL) Fsum->prepare_nextrun();
		
		for (i=0; i < samples; i++) {
			e = G->randomedge();
			if (Fsum != NULL) Fsum->add(e,i,EDGESAMPLING,0); // compare with the optimal + indepedent
		}
		
		if (Fsum != NULL) Fsum->consolidate();
	}
	return err;
  }
  
  
};

class GraphStatistics {
	private:
		Graph *G;
		int32_t flags[256];
		
	public:
		GraphStatistics(Graph *pG) {
			G = pG;
			memset(flags,0,sizeof(flags));
		}
		
		void flood_flag(vertex *u,_FLAGTYPE f) {
			int32_t i;
			vertex *v;
			stack<vertex *> s;
			
			s.push(u);
			while (!s.empty()) { 
				u = s.top();
				s.pop();
				G->setflag(u,f);
				for (i=0; i < u->deg; i++) {
					v = G->getneighbor(u,i);
					if (G->flag(v) == 0) {
						s.push(v);
					}
					else {
						if (G->flag(v) != f) {
							cerr << "Hit flag "<<(int)G->flag(v)<< "  using flag "<<(int)f<<endl;
						}
					}
				}
			}
		}
		
		void flag_connected_components() {
			vertex *v = NULL;
			_FLAGTYPE flag;
			
			G->zeroflags();
			G->vertex_iterator_start();
			flag = 1;
			while ((v = G->next_vertex()) != NULL) {
				if (G->flag(v) == 0) {
					flood_flag(v,flag);
					if (flag < 255) flag++;
				}
			}
		}
		
		void print_connected() {
			int i;
			vertex *v = NULL;
			
			memset(flags,0,sizeof(flags));
			G->vertex_iterator_start();
			while ((v = G->next_vertex()) != NULL) {
				flags[G->flag(v)]++;
			}
			for (i=0; i < 256; i++) {
				if (flags[i] != 0) {
					cerr << "flag["<<i<<"]="<<(int)flags[i]<<endl;
				}
			}
		}
		
		inline int32_t count_flag(_FLAGTYPE flag) {
			return flags[flag];
		}
		
		_FLAGTYPE largest_componentflag() {
			vertex *v = NULL;
			_FLAGTYPE flag=0;
			int32_t maxcount = 0;
			memset(flags,0,sizeof(flags));
			G->vertex_iterator_start();
			while ((v = G->next_vertex()) != NULL) {
				flags[G->flag(v)]++;
				if (G->flag(v) < 255) {
					if (flags[G->flag(v)] > maxcount) {
						maxcount = flags[G->flag(v)];
						flag = G->flag(v);
					}
				}
			}
			return flag;	
		}
		
		void graph_F(F *Fsum) {
			edge *e;
			
			if (Fsum != NULL) Fsum->reset();
			
			if (Fsum != NULL) Fsum->prepare_nextrun();
			
			G->edge_iterator_start();
			while ((e = G->next_edge()) != NULL) {
				if (Fsum != NULL) Fsum->add(e, 0, EDGESAMPLING, 1.0);
			}
			if (Fsum != NULL) Fsum->consolidate_TrueSum();
		}
		
};


class Fbase : public F {
	protected:
		
		_FLOAT *val;
		_FLOAT *X;
		_FLOAT *VARTX;
		int32_t sizeofEprime;
		int32_t samplepoints;
		F *TrueSum;
	public:

		
		Fbase(Graph *pG, F *pTrueSum) : F(pG) {
			val = (_FLOAT *) zalloc((G()->MAXDEGREE()+1)*sizeof(_FLOAT));
			X = (_FLOAT *) zalloc((G()->MAXDEGREE()+1)*sizeof(_FLOAT));
			VARTX = (_FLOAT *) zalloc((G()->MAXDEGREE()+1)*sizeof(_FLOAT));
			sizeofEprime = 0;
			samplepoints = 0;
			TrueSum = pTrueSum;
		}
		
		virtual ~Fbase() {
			free(val);
			free(X);
			free(VARTX);
		}
		
		virtual inline _FLOAT g(edge *e) {
			return (((_FLOAT)1.0)/G()->degree(e->v));
		}
		
		virtual void reset() {
			memset(X,0,(G()->MAXDEGREE()+1)*sizeof(_FLOAT));
			memset(VARTX,0,(G()->MAXDEGREE()+1)*sizeof(_FLOAT));
			prepare_nextrun();
			samplepoints = 0;
		}
		
		virtual void prepare_nextrun() {
			memset(val,0,(G()->MAXDEGREE()+1)*sizeof(_FLOAT));
			sizeofEprime = 0;
		}
		
/*		virtual void add(edge *e, int32_t ) {
			val[G()->degree(e->v)] += g_v(e);
			sizeofEprime++;
		}
		*/
		virtual void add(edge *e, int32_t , int8_t flag, _FLOAT ) {
			if (flag != VERTEXSAMPLING) {
				val[G()->degree(e->v)] += g(e);
				sizeofEprime++;
			}
		}
		
		virtual void consolidate_TrueSum() {
		}
		
		virtual  _FLOAT* trueSum() {
			return val;
		}
		
		virtual void consolidate() {
			int32_t i;
			_FLOAT sumg;
			_FLOAT TrueSumg;
			
/*			TrueSumg = 0.0;
			for (i =1; i <= G()->MAXDEGREE(); i++) {
				TrueSumg += TrueSum->trueSum()[i];
			}*/
			TrueSumg = normalize_TrueSum();
			sumg = normalize_val(val);
			for (i =0; i <= G()->MAXDEGREE(); i++) {
				X[i] += val[i]/sumg;
				VARTX[i] += (val[i]/sumg -  TrueSum->trueSum()[i]/TrueSumg) * (val[i]/sumg -  TrueSum->trueSum()[i]/TrueSumg);
			}
			samplepoints++;
		}
		
		virtual _FLOAT normalize_val(_FLOAT *) {
			return ((_FLOAT)sizeofEprime);
		}
		
		virtual _FLOAT normalize_TrueSum() {
			return TrueSum->normalize_val(NULL);
		}
		
		virtual void print(const char *preamble) {
			int32_t i;
			_FLOAT TrueSumg;
			
			
			TrueSumg = normalize_TrueSum();
			cerr << "#" << preamble << "\tE[\\hat{X}]=\\sum_{k=1}^K\\hat{X}_k/K"<< endl;
			cerr << "#" << preamble << "\ti\tE[\\hat{X}]\t\tE[X]\t\t1-E[\\hat{X}]/E[X]\tsqrt(E[(\\hat{X}-E[X])^2])"<< endl;
			for (i =0; i <= G()->MAXDEGREE(); i++) {
				if (X[i] > 0) 
					cerr << preamble << "\t" << i << "\t" << X[i]/samplepoints << "\t" << (TrueSum->trueSum()[i]/TrueSumg) << "\t" << 1.0-(X[i]/samplepoints)/(TrueSum->trueSum()[i]/TrueSumg)<< "\t" << sqrt(VARTX[i]/samplepoints) << endl;
			}
			cerr << endl;
		}
};


class FBiasedDegDist : public Fbase {
	
	public:
		
	FBiasedDegDist(Graph *pG, F *pTrueSum) : Fbase(pG,pTrueSum) {
	}
	
	virtual void add(edge *e, int32_t , int8_t flag, _FLOAT ) {
		if (flag != VERTEXSAMPLING) {
			val[G()->indegree(e->v)] += 1.0/((_FLOAT)G()->degree(e->v));
			sizeofEprime++;
		}
	}
	
	virtual _FLOAT normalize_TrueSum() {
		int32_t i;
		_FLOAT TrueSumg;
				
		TrueSumg = 0.0;
		for (i =0; i <= G()->MAXDEGREE(); i++) {
			TrueSumg += TrueSum->trueSum()[i];
		}
		return TrueSumg;
	}

	virtual _FLOAT normalize_val(_FLOAT *fvalue) {
		_FLOAT C;
		int32_t i;
		
		C = 0.0;
		for (i =0; i <= G()->MAXDEGREE(); i++) {
			C += fvalue[i];
		}
		return C;
	}
	
};



class FClusteringCoef : public F {
	
	public:
		
		_FLOAT C, Norm, Cavg, CMSE;
		F *TrueSum;
		int32_t runs;
		
		FClusteringCoef(Graph *pG, F *pTrueSum) : F(pG) {
			C = Norm = Cavg = CMSE = 0;
			runs = 0;
			TrueSum = pTrueSum;
		}
		
		virtual void reset() {
			C = Norm = Cavg = CMSE = 0;			
			runs = 0;
		}
		
		virtual void prepare_nextrun() {
			C = Norm = 0;
		}
		
		virtual _FLOAT normalize_val(_FLOAT *) {
			return Norm;
		}
		
		virtual _FLOAT normalize_TrueSum() {
			return Norm;
		}
		
		virtual  _FLOAT* trueSum() {
			return &C;
		}
		
		virtual void consolidate() {
			runs++;
			Cavg += C/normalize_val(NULL);
			CMSE += ((*TrueSum->trueSum())/TrueSum->normalize_TrueSum() - C/normalize_val(NULL)) * ((*TrueSum->trueSum())/TrueSum->normalize_TrueSum() - C/normalize_val(NULL));
		}
		
		virtual void consolidate_TrueSum() {
			Cavg = C/normalize_val(NULL);
		}
		
		virtual void print(const char *Type, F *) {
			cerr << "Clustering Coef\tavg\tNMSE"<< endl;
			if (TrueSum != NULL) {
				cerr << Type << "\t" << Cavg/runs << "\t" << sqrt(CMSE/runs)/((*TrueSum->trueSum())/TrueSum->normalize_TrueSum()) << endl;
			} else {
				cerr << Type << "\t" << Cavg << "\t" << endl;
			}
		}
		
		
		
		_FLOAT c(vertex *u, vertex *v) {
			int32_t i,j;
            vertex *t,*w;
			int32_t common_neighbors;
			
			if ((G()->degree(u) < 2) || (G()->degree(v) < 2))  {
//				cerr << "common_neighbors " << u->id << "," << v->id << " : " << 0<<endl;
				return 0.0;
			}
			
			common_neighbors = 0;
			for  (i=0; i < G()->degree(u) ; i++) {
				t = G()->getneighbor(u,i);
				if (t != NULL) {
					for (j=0; j < G()->degree(v); j++) {
						w = G()->getneighbor(v,j);
						if (w != NULL) {
							if (w->id != u->id) {
                                if (w->id > t->id) break;
                                if (w->id == t->id) common_neighbors++;
							}
						}
					}
				}
			}
//			cerr << "common_neighbors " << u->id << "," << v->id << " : " << common_neighbors<<endl;
			return (2*((_FLOAT)common_neighbors))/(G()->degree(u) + G()->degree(v) - 2);
		}
		
		virtual void add(edge *e, int32_t , int8_t flag, _FLOAT ) {
			_FLOAT f;
			if (flag != VERTEXSAMPLING) {
				if (G()->degree(e->v) >= 2) {
					f = c(G()->getvertex(e->u),G()->getvertex(e->v))/((_FLOAT)G()->degree(e->v));
					C += f;
					Norm += 1.0/((_FLOAT)G()->degree(e->v));
				}
			}
		}
		
};


class FAssortativity : public F {
	
	public:
		
		FAssortativity(Graph *pG) : F(pG) {
		}
		
		virtual void prepare_nextrun() {
			cerr << "NewRunBegin" << endl;
		}
			
		virtual void add(edge *e, int32_t , int8_t flag, _FLOAT ) {
			if (flag != VERTEXSAMPLING) {
				if (e->p > 0) {
					cerr << G()->indegree(e->u) << "," << G()->outdegree(e->u) << "("<< e->u <<")" << "\t" << G()->indegree(e->v) << "," << G()->outdegree(e->v) << "("<< e->v <<")" << "\t" << (int)G()->flag(e->u) << "\t" << (int)G()->flag(e->v) << endl;
				}
			}
		}

		virtual void consolidate_TrueSum() {
			cerr << "NewRunEnd" << endl;
		}

		virtual void consolidate() {
			cerr << "NewRunEnd" << endl;
		}
};

// TODO: assortativity for general labels


int32_t f(int32_t id){
	if (id % 2 == 0) {
		return 1;
	} else {
		return 0;
	}
}

inline _FLOAT sqr(_FLOAT f) {
	return (f*f);
}



#define BASEMULTRUNS 16
#define VARIATIONDISTANCE false


int main(int argc, char *argv[]) {
	Graph *G;
	GraphStatistics *GS;
	RandomEdgeSampling *RE;
	FrontierSampling *FS;
	RandomWalkSampling *RW;
	//Fdegfreq *Fsum;
	FClusteringCoef *Fsum;
	FClusteringCoef *FTrueSum;
	
	int64_t semiruns;
	int64_t samplededges;
	_FLOAT p;
	int64_t steps_samples, walkers;
	int8_t largcomp;
	
	int32_t flag;
	
	char *filename;  
	
	int64_t RUNS=2;
	
	if (argc != 6) {
		cerr << "Missing parameters <filename> <runs>.\n" <<
		"   <filename>: trace file name\n" <<
		"   <steps>: number of total samples ((steps-walkers) in random walks)\n" <<
		"   <walkers>: number of walkers in random walks)\n" <<
		"   <largestcomponent>: 1-just largest component,  0 - all graph)\n" <<
		"   <runs>: number of runs\n" << endl;
		exit(1);
	}
	filename = argv[1];
	steps_samples = atoi(argv[2]);
	walkers = atoi(argv[3]);
	largcomp = atoi(argv[4]);
	RUNS = atoi(argv[5]);
 	
	cerr.precision( 15 );
	cerr.setf(ios::scientific,ios::floatfield);
	
	cerr << "#Filename: "<<filename<<", <steps> = "<<steps_samples<<", <walkers> = "<<walkers<<", runs = "<<RUNS<<endl;
	
	cerr << "#Starting reading graph file...\n";
	G = new Graph(filename);
	cerr << "#Finished reading graph file...\n";
	//   cerr << "#Number of edges = " << G->no_edges() << endl;
	//   cerr << "#Number of vertices = " << G->size() << endl;
	
	GS = new GraphStatistics(G);
	GS->flag_connected_components();
	flag = GS->largest_componentflag();
	cerr << "#Largest component has flag = "<< (int)flag <<" and "<< GS->count_flag(flag) << " nodes. The whole graph has " << G->size() << " nodes"<< endl;
	if (largcomp == 1) {
		G->trimGraphDifferent(flag);
		cerr << "# Trimming graph "<<endl;
	}
	
	G->zeroflags();
	// estimates the distribution from random sampling
	RE = new RandomEdgeSampling(G);
	FS = new FrontierSampling(G);
	RW = new RandomWalkSampling(G);
	//Fsum = new Fdegfreq(G);
	FTrueSum = new FClusteringCoef(G,NULL);
	Fsum = new FClusteringCoef(G,FTrueSum);
	GS->graph_F(FTrueSum);
	FTrueSum->print("AllEdges",NULL);
//		for (h = 0; h < (int64_t)(sizeof(Nvalues)/sizeof(int64_t)); h++) {
		//semiruns = (Nvalues[h]*((int64_t)G->no_edges()))*RUNS;
	semiruns = RUNS;
	cerr << "FS: computing f(G) with ("<<semiruns<<" runs,"<<steps_samples<<" steps,"<< walkers <<" walkers)"<<endl;	
	p = FS->runsampled(steps_samples,walkers,semiruns,false,Fsum,&samplededges);
	Fsum->print("FS",NULL);
	cerr.flush();
	cerr << "RWMW: computing f(G) with ("<<semiruns<<" runs,"<<steps_samples<<" steps,"<< walkers <<" walkers)"<<endl;	
	p = RW->runsampled(steps_samples,walkers,semiruns,false,Fsum,&samplededges);
	Fsum->print("RWMW",NULL);
	cerr.flush();
	cerr << "RWSW: computing f(G) with ("<<semiruns<<" runs,"<<steps_samples<<" steps, 1 walker)"<<endl;	
	p = RW->runsampled(steps_samples,1,semiruns,false,Fsum,&samplededges);
	Fsum->print("RWSW",NULL);
	cerr.flush();
	cerr << "RE: computing f(G) with ("<<semiruns<<" runs,"<<steps_samples<<" samples)"<<endl;	
	p = RE->runsampled(steps_samples,semiruns,false,Fsum);
	Fsum->print("RE",NULL);
	cerr.flush();
		
/*			USB->errorSnowballSampling(4,1,Fsum);
		Fsum->print("SB ",FTrueSum);*/
		//		}
	delete Fsum;
	delete FTrueSum;
	
	delete RE;
	delete FS;
	delete RW;
	if (GS != NULL) delete GS;
	delete G;
	
	return 0;
}

